#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt.async_support
import pytest

import trading_backend.exchanges as exchanges
import tests.util.create_order_tests as create_order_tests
from tests import bybit_exchange


def test_get_name(bybit_exchange):
    assert exchanges.Bybit(bybit_exchange).get_name() == ccxt.async_support.bybit().id.lower()


@pytest.mark.asyncio
async def test_future_orders_parameters(bybit_exchange):
    exchange = exchanges.Bybit(bybit_exchange)
    await create_order_tests.create_order_mocked_test_args(
        exchange,
        symbol="BTC/USDT:USDT",
        exchange_private_post_order_method_name="privatePostPrivateLinearStopOrderCreate",
        exchange_request_referral_key=exchange.HEADER_SPOT_KEY,
        should_contains=False)


@pytest.mark.asyncio
async def test_future_orders_parameters(bybit_exchange):
    bybit_exchange.exchange_manager.is_future = True
    exchange = exchanges.Bybit(bybit_exchange)
    assert exchange.get_headers() == {exchange.HEADER_FUTURE_KEY: exchange._get_id()}
    await create_order_tests.exchange_requests_contains_headers_test(
        exchange,
        exchange_header_referral_key=exchange.HEADER_FUTURE_KEY,
        exchange_header_referral_value=exchange._get_id())



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt
import aiohttp.streams

import trading_backend.exchanges as exchanges
import trading_backend.errors
import trading_backend.enums

class Binance(exchanges.Exchange):
    LEGACY_SPOT_ID = "T9698EB7"
    SPOT_ID = "HR452G85"
    MARGIN_ID = None
    LEGACY_FUTURE_ID = "uquVg2pc"
    FUTURE_ID = "uquVg2pc"
    LEGACY_REF_ID = "135007948"
    REF_ID = "528112221"
    IS_SPONSORING = True

    @classmethod
    def get_name(cls):
        return 'binance'

    def _get_order_custom_id(self):
        return f"x-{self._get_id()}{self._exchange.connector.client.uuid22()}"

    @classmethod
    def use_legacy_ids(cls):
        cls.SPOT_ID = cls.LEGACY_SPOT_ID
        cls.FUTURE_ID = cls.LEGACY_FUTURE_ID

    def _get_broker_dict(self, current_broker_params):
        current_broker_params.update({
            'spot': f"x-{self.__class__.SPOT_ID}",
            'swap': f"x-{self.__class__.FUTURE_ID}",  # used in stop orders
            'future': f"x-{self.__class__.FUTURE_ID}",
        })
        return current_broker_params

    def get_orders_parameters(self, params=None) -> dict:
        current_broker_params = self._exchange.connector.client.options.get("broker", {})
        if current_broker_params != self._get_broker_dict(current_broker_params):
            self._exchange.connector.client.options["broker"] = self._get_broker_dict(current_broker_params)
        return super().get_orders_parameters(params)

    async def _ensure_broker_status(self):
        try:
            if self._exchange.exchange_manager.is_future:
                # no way to check on futures accounts (raising AgentCode is not exist)
                return f"Broker rebate can't be checked when trading futures."
            details = await self._get_account_referral_details()
            if not details.get("rebateWorking", False):
                if (ref_id := details.get("referrerId", None)) and ref_id == self.LEGACY_REF_ID:
                    self.use_legacy_ids()
                    details = await self._get_account_referral_details()
                    if details.get("rebateWorking", False):
                        return "Using legacy broker id"
                return f"Broker rebate not working: {details}"
            return f"Broker rebate is enabled."
        except Exception as err:
            return f"Broker rebate check error: {err}"

    async def _get_account_referral_details(self) -> dict:
        return await self._exchange.connector.client.sapi_get_apireferral_ifnewuser(
            params={
                "apiAgentCode": self._get_id()
            }
        )

    async def _get_api_key_rights(self) -> list[trading_backend.enums.APIKeyRights]:
        try:
            restrictions = await self._exchange.connector.client.sapi_get_account_apirestrictions()
        except ValueError as err:
            raise ccxt.AuthenticationError(f"Invalid key format ({err})")
        except ccxt.NotSupported:
            if self._exchange.exchange_manager.is_sandboxed:
                # API not available on testnet: return all rights
                return [
                    trading_backend.enums.APIKeyRights.READING,
                    trading_backend.enums.APIKeyRights.SPOT_TRADING,
                    trading_backend.enums.APIKeyRights.FUTURES_TRADING,
                    trading_backend.enums.APIKeyRights.MARGIN_TRADING,
                ]
            raise
        rights = []
        if restrictions.get('enableReading'):
            rights.append(trading_backend.enums.APIKeyRights.READING)
        if restrictions.get('enableSpotAndMarginTrading'):
            rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
            rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
        if restrictions.get('enableFutures'):
            rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        if restrictions.get('enableWithdrawals'):
            rights.append(trading_backend.enums.APIKeyRights.WITHDRAWALS)
        return rights

    async def _inner_is_valid_account(self) -> (bool, str):
        details = None
        try:
            if self._exchange.exchange_manager.is_future and not self._exchange.exchange_manager.is_sandboxed:
                # no way to check on futures accounts (raising AgentCode is not exist)
                return True, None
            details = await self._get_account_referral_details()
            if not details.get("rebateWorking", False):
                ref_id = details.get("referrerId", None)
                if ref_id is not None:
                    return False, f"This account has a referral id equal to {ref_id} " \
                                  f"which is incompatible ({self.REF_ID} as referral id or no referral id is required)"
                return False, f"This account is incompatible, details: {details}. Please report this message to " \
                              f"admins for investigation. " \
                              f"An account with {self.REF_ID} as referral id or no referral id is required."
            if not details.get("ifNewUser", False):
                return False, "Binance requires accounts that were created after july 1st 2021, " \
                              "this account is too old."
        except ccxt.NotSupported:
            if self._exchange.exchange_manager.is_sandboxed:
                # API not available on testnet
                try:
                    await self._exchange.get_balance()
                    return True, None
                except Exception as err:
                    raise ccxt.AuthenticationError(f"Invalid auth details ({err})")
            raise
        except ValueError as err:
            raise ccxt.AuthenticationError(f"Invalid key format ({err})")
        except ccxt.ExchangeError as err:
            if "AgentCode is not exist" in str(err):
                # err: BadRequest('binance {"code":-9000,"msg":"apiAgentCode is not exist"}')
                raise trading_backend.errors.InvalidIdError(
                    f"{self.__class__.__name__} Agent code error (configured local id) is invalid. Please update it ({err})"
                )
            raise err
        except AttributeError:
            if isinstance(details, aiohttp.streams.StreamReader):
                return False, "Error when fetching exchange data (unreadable response)"
            return False, "Invalid request parameters"
        return True, None



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy

import pytest
import mock
import ccxt.async_support
import trading_backend.exchanges as exchanges
import trading_backend.enums
import trading_backend
import tests.util.create_order_tests as create_order_tests
import tests.util.account_tests as account_tests
from tests import binance_exchange


def test_get_name(binance_exchange):
    assert exchanges.Binance(binance_exchange).get_name() == ccxt.async_support.binance().id.lower()


@pytest.mark.asyncio
async def test_get_orders_parameters(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    await create_order_tests.create_order_mocked_test_args(
        exchange,
        exchange_private_post_order_method_name="privatePostOrder",
        exchange_request_referral_key="newClientOrderId",
        should_contains=True
    )


@pytest.mark.asyncio
async def test_get_api_key_rights(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    with mock.patch.object(
        exchange._exchange.connector.client, "sapi_get_account_apirestrictions",
        mock.AsyncMock(return_value={"enableReading": True, "enableSpotAndMarginTrading": False, "enableFutures": True, "enableWithdrawals": False})
    ) as sapi_get_account_apirestrictions_mock:
        assert await exchange._get_api_key_rights() == [
            trading_backend.enums.APIKeyRights.READING,
            trading_backend.enums.APIKeyRights.FUTURES_TRADING,
        ]
        sapi_get_account_apirestrictions_mock.assert_awaited_once()
    with mock.patch.object(
        exchange._exchange.connector.client, "sapi_get_account_apirestrictions",
        mock.AsyncMock(return_value={"enableReading": True, "enableSpotAndMarginTrading": True, "enableWithdrawals": True})
    ) as sapi_get_account_apirestrictions_mock:
        assert await exchange._get_api_key_rights() == [
            trading_backend.enums.APIKeyRights.READING,
            trading_backend.enums.APIKeyRights.SPOT_TRADING,
            trading_backend.enums.APIKeyRights.MARGIN_TRADING,
            trading_backend.enums.APIKeyRights.WITHDRAWALS
        ]
        sapi_get_account_apirestrictions_mock.assert_awaited_once()


@pytest.mark.asyncio
async def test_is_valid_account(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    params = {"apiAgentCode": exchange._get_id()}
    with pytest.raises(trading_backend.ExchangeAuthError):
        await exchange.is_valid_account()

    with mock.patch.object(
            exchange, "_get_api_key_rights", mock.AsyncMock(return_value=[trading_backend.enums.APIKeyRights.READING])
    ) as _get_api_key_rights_mock:
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) \
                as sapi_get_apireferral_ifnewuser_mock:
            with pytest.raises(trading_backend.errors.APIKeyPermissionsError):
                assert (await exchange.is_valid_account()) == (True, None)
                sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
                _get_api_key_rights_mock.assert_called_once()

    with mock.patch.object(
            exchange, "_get_api_key_rights", mock.AsyncMock(return_value=[
                trading_backend.enums.APIKeyRights.SPOT_TRADING, trading_backend.enums.APIKeyRights.WITHDRAWALS
            ])
    ) as _get_api_key_rights_mock:
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) \
                as sapi_get_apireferral_ifnewuser_mock:
            with mock.patch.object(exchange, "_allow_withdrawal_right", mock.Mock(return_value=True)) as \
                 _allow_withdrawal_right_mock:
                assert (await exchange.is_valid_account()) == (True, None)
                sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
                _get_api_key_rights_mock.assert_called_once()
                _allow_withdrawal_right_mock.assert_called_once()
            with mock.patch.object(exchange, "_allow_withdrawal_right", mock.Mock(return_value=False)) as \
                 _allow_withdrawal_right_mock:
                with pytest.raises(trading_backend.errors.APIKeyPermissionsError):
                    assert (await exchange.is_valid_account()) == (True, None)
                _allow_withdrawal_right_mock.assert_called_once()

    with mock.patch.object(
            exchange, "_get_api_key_rights", mock.AsyncMock(return_value=[trading_backend.enums.APIKeyRights.SPOT_TRADING])
    ) as _get_api_key_rights_mock:
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": False})) as sapi_get_apireferral_ifnewuser_mock:
            results = await exchange.is_valid_account(always_check_key_rights=False)
            assert results[0] is False
            assert isinstance(results[1], str)
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
            _get_api_key_rights_mock.assert_not_called()
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"ifNewUser": False})) as sapi_get_apireferral_ifnewuser_mock:
            results = await exchange.is_valid_account(always_check_key_rights=True)
            assert results[0] is False
            assert isinstance(results[1], str)
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
            _get_api_key_rights_mock.assert_called_once()
            _get_api_key_rights_mock.reset_mock()
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={})) as sapi_get_apireferral_ifnewuser_mock:
            results = await exchange.is_valid_account()
            assert results[0] is False
            assert isinstance(results[1], str)
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
            _get_api_key_rights_mock.assert_not_called()
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": False, "ifNewUser": True})) \
                as sapi_get_apireferral_ifnewuser_mock:
            results = await exchange.is_valid_account()
            assert results[0] is False
            assert isinstance(results[1], str)
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value=None)) \
                as sapi_get_apireferral_ifnewuser_mock:
            results = await exchange.is_valid_account()
            assert results[0] is False
            assert isinstance(results[1], str)
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
            _get_api_key_rights_mock.assert_not_called()
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) \
                as sapi_get_apireferral_ifnewuser_mock:
            assert (await exchange.is_valid_account()) == (True, None)
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
            _get_api_key_rights_mock.assert_called_once()
            _get_api_key_rights_mock.reset_mock()
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(side_effect=ccxt.async_support.InvalidNonce())) \
                as sapi_get_apireferral_ifnewuser_mock:
            with pytest.raises(trading_backend.TimeSyncError):
                await exchange.is_valid_account()
            sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
            _get_api_key_rights_mock.assert_not_called()


    # futures trading
    exchange._exchange.exchange_manager.is_future = True
    with mock.patch.object(
        exchange, "_get_api_key_rights", mock.AsyncMock(return_value=[
            trading_backend.enums.APIKeyRights.FUTURES_TRADING, trading_backend.enums.APIKeyRights.WITHDRAWALS
        ])
    ) as _get_api_key_rights_mock:
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) \
                as sapi_get_apireferral_ifnewuser_mock:
            with mock.patch.object(exchange, "_allow_withdrawal_right", mock.Mock(return_value=True)) as \
                 _allow_withdrawal_right_mock:
                assert (await exchange.is_valid_account()) == (True, None)
                sapi_get_apireferral_ifnewuser_mock.assert_not_called()
                _get_api_key_rights_mock.assert_called_once()
                _allow_withdrawal_right_mock.assert_called_once()
            with mock.patch.object(exchange, "_allow_withdrawal_right", mock.Mock(return_value=False)) as \
                 _allow_withdrawal_right_mock:
                with pytest.raises(trading_backend.errors.APIKeyPermissionsError):
                    assert (await exchange.is_valid_account()) == (True, None)
                _allow_withdrawal_right_mock.assert_called_once()
    with mock.patch.object(
        exchange, "_get_api_key_rights", mock.AsyncMock(return_value=[
            trading_backend.enums.APIKeyRights.SPOT_TRADING, trading_backend.enums.APIKeyRights.WITHDRAWALS
        ])
    ) as _get_api_key_rights_mock:
        with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                               mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) \
                as sapi_get_apireferral_ifnewuser_mock:
            with mock.patch.object(exchange, "_allow_withdrawal_right", mock.Mock(return_value=True)) as \
                 _allow_withdrawal_right_mock:
                with pytest.raises(trading_backend.errors.APIKeyPermissionsError):
                    # futures trading permissions are required
                    assert (await exchange.is_valid_account()) == (True, None)
                sapi_get_apireferral_ifnewuser_mock.assert_not_called()
                _get_api_key_rights_mock.assert_called_once()
                _allow_withdrawal_right_mock.assert_not_called()


@pytest.mark.asyncio
async def test_ensure_broker_status(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    with mock.patch.object(exchange, "use_legacy_ids", mock.Mock()) as use_legacy_ids_mock:
        with mock.patch.object(exchange, "_get_account_referral_details", mock.AsyncMock(side_effect=RuntimeError)) \
                as _get_account_referral_details_mock:
            assert "check error" in await exchange._ensure_broker_status()
            _get_account_referral_details_mock.assert_awaited_once()
        with mock.patch.object(exchange, "_get_account_referral_details",
                               mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) as \
                _get_account_referral_details_mock:
            assert "is enabled" in await exchange._ensure_broker_status()
            _get_account_referral_details_mock.assert_awaited_once()
            use_legacy_ids_mock.assert_not_called()
        with mock.patch.object(exchange, "_get_account_referral_details",
                               mock.AsyncMock(return_value={"rebateWorking": False, "ifNewUser": True})) as \
                _get_account_referral_details_mock:
            assert "not working" in await exchange._ensure_broker_status()
            _get_account_referral_details_mock.assert_awaited_once()
            use_legacy_ids_mock.assert_not_called()
        with mock.patch.object(exchange, "_get_account_referral_details",
                               mock.AsyncMock(return_value={"rebateWorking": False, "ifNewUser": True,
                                                            "referrerId": exchange.LEGACY_REF_ID})) as \
                _get_account_referral_details_mock:
            assert "not working" in await exchange._ensure_broker_status()
            assert _get_account_referral_details_mock.await_count == 2
            use_legacy_ids_mock.assert_called_once()
            use_legacy_ids_mock.reset_mock()

        async def _dep_on_id_mock():
            if use_legacy_ids_mock.call_count != 0:
                return {"rebateWorking": True, "ifNewUser": True, "referrerId": exchange.LEGACY_REF_ID}
            return {"rebateWorking": False, "ifNewUser": True, "referrerId": exchange.LEGACY_REF_ID}
        with mock.patch.object(exchange, "_get_account_referral_details",
                               mock.AsyncMock(side_effect=_dep_on_id_mock)) as \
                _get_account_referral_details_mock:
            assert "legacy broker id" in await exchange._ensure_broker_status()
            assert _get_account_referral_details_mock.await_count == 2
            use_legacy_ids_mock.assert_called_once()
            use_legacy_ids_mock.reset_mock()

def test_use_legacy_ids(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    origin_spot_id = copy.copy(exchange.SPOT_ID)
    origin_legacy_spot_id = copy.copy(exchange.LEGACY_SPOT_ID)
    origin_future_id = copy.copy(exchange.FUTURE_ID)
    origin_legacy_future_id = copy.copy(exchange.LEGACY_FUTURE_ID)
    assert exchange.SPOT_ID == origin_spot_id
    assert exchange._get_id() == origin_spot_id
    assert exchange.FUTURE_ID == origin_future_id
    assert origin_legacy_spot_id != origin_spot_id
    exchange.use_legacy_ids()
    assert exchange.SPOT_ID == origin_legacy_spot_id
    assert exchange._get_id() == origin_legacy_spot_id
    assert exchange.FUTURE_ID == origin_legacy_future_id


@pytest.mark.asyncio
async def test_invalid_api_key(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    await account_tests.check_invalid_account(exchange)


@pytest.mark.asyncio
async def test_invalid_api_key_get_api_key_rights(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    await account_tests.check_invalid_account_keys_rights(exchange)



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt

import trading_backend.exchanges as exchanges
import trading_backend.enums


class Bingx(exchanges.Exchange):
    SPOT_ID = "OctoBot"
    FUTURE_ID = "OctoBot"
    IS_SPONSORING = True

    @classmethod
    def get_name(cls):
        return 'bingx'

    async def _inner_cancel_order(self):
        # use client api to avoid any ccxt call wrapping and error handling
        try:
            await self._exchange.connector.client.cancel_order("12345", symbol=self._get_symbol())
        except ccxt.ExchangeError as err:
            # ('bingx {"code":100413,"msg":"Incorrect apiKey","timestamp":1718551786654}',)
            if "Incorrect apiKey".lower() in str(err).lower():
                # error is not caught by ccxt as such
                raise ccxt.AuthenticationError(f"Invalid key format ({err})")
            raise

    async def _get_api_key_rights(self) -> list[trading_backend.enums.APIKeyRights]:
        # It is currently impossible to fetch api key permissions: try to cancel an imaginary order,
        # if a permission error is raised instead of a cancel fail, then trading permissions are missing.
        # updated: 24/01/2024
        return await self._get_api_key_rights_using_order()

        # todo use this when available (current issue is that ccxt bingx sign() requires "v1" to be the second value of
        # the path and it's the 1st in /openApi/v1/account/apiRestrictions which crashes in  implode_hostname

        # https://bingx-api.github.io/docs/#/en-us/common/permission-interface.html#Query%20user%20API%20Key%20permissions
        # GET /openApi/v1/account/apiRestrictions
        # not on ccxt yet
        # restrictions = await self._exchange.connector.client.accountV1PrivateGetapiRestrictions(self._exchange.connector.client)
        # rights = []
        # if restrictions.get('enableReading'):
        #     rights.append(trading_backend.enums.APIKeyRights.READING)
        # if restrictions.get('enableSpotAndMarginTrading'):
        #     rights.append(trading_backend.enums.APIKeyRights.SPOT_TRADING)
        #     rights.append(trading_backend.enums.APIKeyRights.MARGIN_TRADING)
        # if restrictions.get('enableFutures'):
        #     rights.append(trading_backend.enums.APIKeyRights.FUTURES_TRADING)
        # if restrictions.get('permitsUniversalTransfer'):
        #     rights.append(trading_backend.enums.APIKeyRights.WITHDRAWALS)
        # return rights

    def get_orders_parameters(self, params=None) -> dict:
        if self._exchange.connector.client.options.get("broker", None) != self._get_id():
            self._exchange.connector.client.options["broker"] = self._get_id()
        return super().get_orders_parameters(params)



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import ccxt.async_support
import trading_backend.exchanges as exchanges
import trading_backend.enums
import trading_backend
from tests import coinbase_exchange


def test_get_name(coinbase_exchange):
    assert exchanges.Coinbase(coinbase_exchange).get_name() == ccxt.async_support.coinbase().id.lower()


@pytest.mark.asyncio
async def test_get_orders_parameters(coinbase_exchange):
    exchange = exchanges.Coinbase(coinbase_exchange)
    assert exchange.get_orders_parameters({}) == {}


@pytest.mark.asyncio
async def test_inner_is_valid_account(coinbase_exchange):
    exchange = exchanges.Coinbase(coinbase_exchange)
    assert await exchange._inner_is_valid_account() == (False, await exchange._ensure_broker_status())


@pytest.mark.asyncio
async def test_invalid_api_key(coinbase_exchange):
    # _inner_is_valid_account is not implemented on coinbase
    pass


@pytest.mark.asyncio
async def test_invalid_api_key_get_api_key_rights(coinbase_exchange):
    # API keys used as a base for these tests are deleted from Coinbase
    exchange = exchanges.Coinbase(coinbase_exchange)
    with pytest.raises(ccxt.AuthenticationError):
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # binascii.Error turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "organizations/1e7665a0-440b-495f-8a49-5365841e196e/apiKeys/b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "AwEHoUQDQgAErQXGFQUMqPT5fQUVcCchCaomapu0y952"
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # AssertionError turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "organizations/1e7665a0-440b-495f-8a49-5365841e196e/apiKeys/b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "dd"
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # IndexError turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "organizations/1e7665a0-440b-495f-8a49-5365841e196e/apiKeys/b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEINpzmww4rUF+EeSMDnBqd3oXTxm2Q76MUwwsLw8Vjo+poAoGCCqGSM49\nBk6cNPNp4fH0NwneES1HNpJ0aEx+VYRhcg=="
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # binascii.Error turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "-495f-8a49-5365841e196e/apiKeys/b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEIK/XmQtNSTD2pqrhyZFvBuExnlyLOPcDzT+fsm1sq3ZCoAoGCCqGSM49\nAwEHoUQDQgAErQXGFQUMqPT5fQUVcCchCaomapu0y952+XgveL2QjgghGCeFbLfR\nTSg2RgUUtGbG3TIBEomwzbRAOEeYdjK06w"
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # binascii.Error turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEIK/XmQtNSTD2pqrhyZFvBuExnlyLOPcDzT+fsm1sq3ZCoAoGCCqGSM49\nAwEHoUQDQgAErQXGFQUMqPT5fQUVcCchCaomapu0y952+XgveL2QjgghGCeFbLfR\nTSg2RgUUtGbG3TIBEomwzbRAOEeYdjK06w"
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # swapped private and public
        # ccxt.ArgumentsRequired turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "-----BEGIN EC PRIVATE KEY----b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEIK/XmQtNSTD2pqrhyZFvBuExnlyLOPcDzT+fsm1sq3ZCoAoGCCqGSM49\nAwEHoUQDQgAErQXGFQUMqPT5fQUVcCchCaomapu0y952+XgveL2QjgghGCeFbLfR\nTSg2RgUUtGbG3TIBEomwzbRAOEeYdjK06w"
        assert await exchange._get_api_key_rights()
    with pytest.raises(ccxt.AuthenticationError):
        # added chars after private key
        # ccxt.static_dependencies.ecdsa.der.UnexpectedDER turned into ccxt.AuthenticationError
        exchange._exchange.connector.client.apiKey = "organizations/1e7665a0-440b-495f-8a49-5365841e196e/apiKeys/b43c5889-a3b5-4ab5-a606-578b4d74f3db"
        exchange._exchange.connector.client.secret = "-----BEGIN EC PRIVATE KEY-----\nMHcCAQEEIK/XmQtNSTD2pqrhyZFvBuExnlyLOPcDzT+fsm1sq3ZCoAoGCCqGSM49\nAwEHoUQDQgAErQXGFQUMqPT5fQUVcCchCaomapu0y952+XgveL2QjgghGCeFbLfR\nTSg2RgUUtGbG3TIBEomwzbRAOEeYdjK06w" + b'488347'.decode()
        assert await exchange._get_api_key_rights()


# @pytest.mark.asyncio
# async def test_get_api_key_rights(coinbase_exchange):
#     exchange = exchanges.Coinbase(coinbase_exchange)
#     with mock.patch.object(
#         exchange._exchange.connector.client, "v2PrivateGetUserAuth",
#         mock.AsyncMock(return_value={"data": {"scopes": [
#             "rat#view",
#             "rat#trade",
#             "rat#transfer",
#         ]}})
#     ) as v2PrivateGetUserAuth_mock:
#         assert await exchange._get_api_key_rights() == [
#             trading_backend.enums.APIKeyRights.READING,
#             trading_backend.enums.APIKeyRights.SPOT_TRADING,
#             trading_backend.enums.APIKeyRights.MARGIN_TRADING,
#             trading_backend.enums.APIKeyRights.FUTURES_TRADING,
#             trading_backend.enums.APIKeyRights.WITHDRAWALS,
#         ]
#         v2PrivateGetUserAuth_mock.assert_awaited_once()
#     with mock.patch.object(
#         exchange._exchange.connector.client, "v2PrivateGetUserAuth",
#         mock.AsyncMock(return_value={"data": {"scopes": [
#             "rat#view",
#             "rat#trade",
#         ]}})
#     ) as v2PrivateGetUserAuth_mock:
#         assert await exchange._get_api_key_rights() == [
#             trading_backend.enums.APIKeyRights.READING,
#             trading_backend.enums.APIKeyRights.SPOT_TRADING,
#             trading_backend.enums.APIKeyRights.MARGIN_TRADING,
#             trading_backend.enums.APIKeyRights.FUTURES_TRADING,
#         ]
#         v2PrivateGetUserAuth_mock.assert_awaited_once()
#     with mock.patch.object(
#         exchange._exchange.connector.client, "v2PrivateGetUserAuth",
#         mock.AsyncMock(return_value={"data": {"scopes": [
#             "wallet:accounts:read",
#             "wallet:buys:read",
#             "wallet:sells:read",
#             "wallet:orders:read",
#             "wallet:trades:read",
#             "wallet:user:read",
#             "wallet:transactions:read",
#             "wallet:buys:create",
#         ]}})
#     ) as v2PrivateGetUserAuth_mock:
#         assert await exchange._get_api_key_rights() == [
#             trading_backend.enums.APIKeyRights.READING
#         ]
#         v2PrivateGetUserAuth_mock.assert_awaited_once()
#     with mock.patch.object(
#         exchange._exchange.connector.client, "v2PrivateGetUserAuth",
#         mock.AsyncMock(return_value={"data": {"scopes": [
#             "wallet:accounts:read",
#             "wallet:buys:read",
#             "wallet:sells:read",
#             "wallet:orders:read",
#             "wallet:trades:read",
#             "wallet:user:read",
#             "wallet:transactions:read",
#             "wallet:buys:create",
#             "wallet:sells:create",
#             "wallet:withdrawals:create",
#         ]}})
#     ) as v2PrivateGetUserAuth_mock:
#         assert await exchange._get_api_key_rights() == [
#             trading_backend.enums.APIKeyRights.READING,
#             trading_backend.enums.APIKeyRights.SPOT_TRADING,
#             trading_backend.enums.APIKeyRights.MARGIN_TRADING,
#             trading_backend.enums.APIKeyRights.FUTURES_TRADING,
#             trading_backend.enums.APIKeyRights.WITHDRAWALS
#         ]
#         v2PrivateGetUserAuth_mock.assert_awaited_once()
